

#include "lobscfg.h"
#include "lobsdims.h"
#include "lobsc.h"


typedef struct lobs_front_end_t *lobs_front_end;

typedef void (*lobs_error_notify_func) (char *file, int line, char *explanation);



/*
  Create a new Lobs (point-in-range tree).
 */
lobs_front_end
lobs_new_instance (lobs_error_notify_func error_notify,
                   int per_leaf
                   );


/*
   Set the maximal number of retrieved objects.
 */
int
lobs_set_find_max_num_objects (lobs_front_end lobs, int find_max_num_objects);


/*
   The function adds an object /object/ to the search tree.
   The bounding-box of the object is specified by the coordinates
   of two points, /ll/ and /ur/.  The coordinates of these points
   must bound the object.

   Example:  The sphere of radius 1, with a center at {-2;3;8}
   can be described by the bounding-box given by the points
   {-3;4;7} and {-1;2;9}.  Note that the lower/upper distinction is
   not necessarily preserved in all coordinate directions for the two
   points.
   */
int                             /* the returned value can be ignored */
lobs_add_object (lobs_front_end lobs,/* front end */
                 void *object,  /* object to insert */
                 REF_LIST(double llx,
                          double lly,
                          double llz), /* the coordinates of the left-lower... */
                 REF_LIST(double urx,
                          double ury,
                          double urz) /* ... and upper-right corner
                                         of the object bounding box */
                 );


/*
   Search the tree.
   The function searches for all objects whose
   bounding-boxes overlap the box specified on input.  The box
   on input may be a single point, of course. 
   */
int /* the returned value can be ignored */
lobs_search (lobs_front_end lobs,/* front end */
             REF_LIST(double llx,
                      double lly,
                      double llz),/* the coordinates of the left-lower */
             REF_LIST(double urx,
                      double ury,
                      double urz)/* ... and upper-right corner
                                    of the box  */
             );

/*
   This function searches the tree for all the objects whose
   bounding-boxes overlap the point specified on input.  The
   function is equivalent to lobs_search (), where the input
   box degenerates to a point.
   */
int
lobs_search_for_boxes_overlapping_pt (lobs_front_end lobs,
                                      REF_LIST(double ptx,
                                               double pty,
                                               double ptz));

/*
   Get the number of objects found.
   Get the number of objects found to overlap a given box (point).
 */
int
lobs_num_of_objects_found (lobs_front_end lobs);

void *
lobs_first_overlapping_object (lobs_front_end lobs);


void *
lobs_next_overlapping_object (lobs_front_end lobs);


/*
   Print the structure of the LOBS tree to a stream specified.
 */
int
lobs_print(lobs_front_end lobs, FILE *fp);

/*
   Free all memory used by the LOBS tree.  The instance of the
   search tree may not be
   used subsequently.
 */
lobs_front_end                  /* returns NULL */
lobs_free_instance (lobs_front_end lobs /* front end to destroy */
                    );

/*
   Recycle instance.
   The stored objects are discarded, and if the implementation supports this,
   the lobs structure returned can still be used by filling in new objects.
   On the other hand, for the malloc-based implementation, this function
   is identical to lobs_free_instance(), and NULL is returned.
 */
lobs_front_end
lobs_recycle_instance (lobs_front_end lobs);


/*
   The function moves an object /object/ stored previously in the search tree.
   The bounding-box of the object is specified by the coordinates
   of two points, /ll/ and /ur/.  The coordinates of these points
   must bound the object.

   Example:  The sphere of radius 1, with a center at  {-2;3;8} 
   can be described by the bounding-box given by the points
    {-3;4;7}  and  {-1;2;9} .  Note that the lower/upper distinction is
   not necessarily preserved in all coordinate directions for the two
   points.
   */
int
lobs_move_object (lobs_front_end lobs,
                  void *object,
                  REF_LIST (double old_llx,
                            double old_lly,
                            double old_llz) ,
                  REF_LIST (double old_urx,
                            double old_ury,
                            double old_urz),
                  REF_LIST (double llx,
                            double lly,
                            double llz) ,
                  REF_LIST (double urx,
                            double ury,
                            double urz)) ;
