#include <stdio.h>
#include <time.h>

#include "randP.h"
#include "timeP.h"
#include "getoptP.h"
#include "g3d.h"



#ifndef TIMING_MODE
#   define TIMING_MODE   TRUE
#endif
#ifndef TEST_DIM
#   define TEST_DIM 3
#endif
#ifndef TEST_BOX
#   define TEST_BOX 0
#endif

#define NOBJS    1000
#define NSRCHS   5000

#if   TEST_DIM == 1
#include "lobs1d.h"
#elif TEST_DIM == 2
#include "lobs2d.h"
#elif TEST_DIM == 3
#include "lobs3d.h"
#endif


typedef struct obj {
#if   TEST_DIM == 1
  int        found_by_lobs_1d;
  int        found_by_brute_1d;
#elif TEST_DIM == 2
  int        found_by_lobs_2d;
  int        found_by_brute_2d;
#elif TEST_DIM == 3
  int        found_by_lobs_3d;
  int        found_by_brute_3d;
#endif
  G3D_box_t  box;
}              obj;

#if   TEST_DIM >= 1
static double xscalelo = -1;
static double xscalehi = 1;
#endif

#if   TEST_DIM >= 2
static double yscalelo = 0.3;
static double yscalehi = 0.9;
#else
static double yscalelo = 0;
static double yscalehi = 0;
#endif

#if   TEST_DIM >= 3
static double zscalelo = 0; 
static double zscalehi = 1; 
#else
static double zscalelo = 0; 
static double zscalehi = 0; 
#endif


static int nobjs       = NOBJS;
static int nsearches   = NSRCHS;
static int pr_scale    = NSRCHS / 10;
static int per_leaf    = 10;
static int inner_loops = 10;


static obj *objs;


static double cumulative_brute_time = 0;
static double cumulative_lobs_time = 0;


static void 
generate(void);
static void
test_lobs(void);
static void
search_brute(G3D_box_t *box);
static void
search_lobs(G3D_box_t *box);
static void
time_lobs(void);
static void 
process_options(int argc, char **argv);

static void
move (void);


int  
main(int argc, char **argv)
{
  double real_time, usr_time, systm_time;
  double uts;
  
  process_options (argc, argv);
  
#if TIMING_MODE
  fprintf (stderr, "Timing %dD library: %d objects, %d searches\n",
           TEST_DIM, nobjs, nsearches * inner_loops);
#else
  fprintf (stderr, "Testing %dD library: %d objects, %d searches\n",
           TEST_DIM, nobjs, nsearches);
#endif
  
  ckit_get_times(&real_time, &usr_time, &systm_time);
  uts = usr_time;
  generate ();
  ckit_get_times(&real_time, &usr_time, &systm_time);
  fprintf (stderr, "Generate () took %g sec\n", usr_time - uts);
  fprintf (stderr, "Generated:\n");
#if TIMING_MODE
  time_lobs ();
#else
  test_lobs ();
#endif
  
  ckit_get_times(&real_time, &usr_time, &systm_time);
  uts = usr_time;
  move ();
  ckit_get_times(&real_time, &usr_time, &systm_time);
  fprintf (stderr, "Move () took %g sec\n", usr_time - uts);
  fprintf (stderr, "Moved:\n");
#if TIMING_MODE
  time_lobs ();
#else
  test_lobs ();
#endif
  

  fprintf (stderr, "Success\n");

  return 0;
}

#define LOBS_NEW_ARRAY(NELT, TYPE) \
                      ((TYPE *) calloc((NELT), (size_t) sizeof(TYPE)))


static int
errexit(char *s)
{
  fprintf(stderr, "Error: %s\n", s);
  exit(1);
  return 1;
}


#if   TEST_DIM == 1
static lobs_1d_front_end lobs1d = NULL;
#elif TEST_DIM == 2
static lobs_2d_front_end lobs2d = NULL;
#elif TEST_DIM == 3
static lobs_3d_front_end lobs3d = NULL;
#endif

/* Ordinary loop */
#define FOR(_Counter_, _Up_)                                      \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)

static void
next_obj(int i);



static void
generate (void)
{
  int i;

  objs = LOBS_NEW_ARRAY(nobjs, obj);
  if (objs == NULL) errexit("malloc failed");

#if   TEST_DIM == 1
  lobs1d = lobs_1d_new_instance (NULL, per_leaf);
#elif TEST_DIM == 2
  lobs2d = lobs_2d_new_instance (NULL, per_leaf);
#elif TEST_DIM == 3
  lobs3d = lobs_3d_new_instance (NULL, per_leaf);
#endif
  
  FOR(i, nobjs) {
    next_obj(i);
#if   TEST_DIM == 1
    lobs_1d_add_object (lobs1d, &objs[i],
                        objs[i].box.lower.x, objs[i].box.upper.x);
#elif TEST_DIM == 2
    lobs_2d_add_object (lobs2d, &objs[i],
                        objs[i].box.lower.x,
                        objs[i].box.lower.y,
                        objs[i].box.upper.x,
                        objs[i].box.upper.y);
#elif TEST_DIM == 3
    lobs_3d_add_object (lobs3d, &objs[i],
                        objs[i].box.lower.x,
                        objs[i].box.lower.y,
                        objs[i].box.lower.z,
                        objs[i].box.upper.x,
                        objs[i].box.upper.y,
                        objs[i].box.upper.z);
#endif
  }
}

static void
move_obj (int i)
{
  G3D_loc_t loc;
  int nn = pow ((double) nobjs, (double) 1/TEST_DIM) + 1;
  double bsize = ((xscalehi - xscalelo) + (yscalehi - yscalelo) + (zscalehi - zscalelo)) / nn;
  double msize = bsize / 10;
  
  loc.x = (objs[i].box.upper.x + objs[i].box.lower.x) / 2 + msize * rand_0_to_1();
  loc.y = (objs[i].box.upper.y + objs[i].box.lower.y) / 2 + msize * rand_0_to_1();
  loc.z = (objs[i].box.upper.z + objs[i].box.lower.z) / 2 + msize * rand_0_to_1();
  G3D_INIT_BOX (objs[i].box, loc);
  G3D_INFLATE_BOX (objs[i].box, bsize);
}

static void
move (void)
{
  int i;
  G3D_box_t oldbox;

  FOR(i, nobjs) {
    oldbox = objs[i].box;
    move_obj (i);
#if   TEST_DIM == 1
    if (!lobs_1d_move_object (lobs1d, &objs[i],
                              oldbox.lower.x, oldbox.upper.x,
                              objs[i].box.lower.x, objs[i].box.upper.x))
      errexit ("move failed");
#elif TEST_DIM == 2
    if (!lobs_2d_move_object (lobs2d, &objs[i],
                              oldbox.lower.x, oldbox.lower.y,
                              oldbox.upper.x, oldbox.upper.y,
                              objs[i].box.lower.x, objs[i].box.lower.y,
                              objs[i].box.upper.x, objs[i].box.upper.y))
      errexit ("move failed");
#elif TEST_DIM == 3
    if (!lobs_3d_move_object (lobs3d, &objs[i],
                              oldbox.lower.x, oldbox.lower.y, oldbox.lower.z,
                              oldbox.upper.x, oldbox.upper.y, oldbox.lower.z,
                              objs[i].box.lower.x, objs[i].box.lower.y, objs[i].box.lower.z,
                              objs[i].box.upper.x, objs[i].box.upper.y, objs[i].box.upper.z))
      errexit ("move failed");
#endif
  }
}



static void
next_obj (int i)
{
  G3D_loc_t loc;
  int nn = pow ((double) nobjs, (double) 1/TEST_DIM) + 1;
  double bsize = ((xscalehi - xscalelo) + (yscalehi - yscalelo) + (zscalehi - zscalelo)) / nn;
  double msize = bsize / 10;
  
#if 0                           /* arbitrary size object */
  loc.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
  loc.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
  loc.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
  G3D_INIT_BOX(objs[i].box, loc);
  loc.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
  loc.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
  loc.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
  G3D_UPDT_BOX(objs[i].box, loc);
#else
  loc.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
  loc.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
  loc.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
  G3D_INIT_BOX(objs[i].box, loc);
  G3D_INFLATE_BOX(objs[i].box, bsize);
#endif
}

static void
compare(G3D_box_t *box);

static void
test_lobs (void)
{
  int i, j;
  G3D_loc_t p;
  G3D_box_t box;

  pr_scale = nsearches / 10;
  FOR(i, nsearches) {
    FOR(j, nobjs) {
#if   TEST_DIM == 1
      objs[j].found_by_lobs_1d  = 0;
      objs[j].found_by_brute_1d = 0;
#elif TEST_DIM == 2
      objs[j].found_by_lobs_2d  = 0;
      objs[j].found_by_brute_2d = 0;
#elif TEST_DIM == 3
      objs[j].found_by_lobs_3d  = 0;
      objs[j].found_by_brute_3d = 0;
#endif
    }
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_INIT_BOX(box, p);
#   if TEST_BOX
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_UPDT_BOX(box, p);
#   endif
    search_lobs (&box);
    search_brute (&box);
    compare(&box);
    if ((i + 1) / pr_scale * pr_scale == (i + 1))
      fprintf(stderr, " done %d of %d\n", (i + 1), nsearches);
  }
}



  


static void
time_lobs(void)
{
  int i, j;
  G3D_loc_t p;
  G3D_box_t box;
  double real_time, usr_time, systm_time;
  double lobs_time, brute_time;

  cumulative_brute_time = 0;
  cumulative_lobs_time = 0;
  FOR(i, nsearches) {
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_INIT_BOX(box, p);
#   if TEST_BOX
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_UPDT_BOX(box, p);
#   endif
    ckit_get_times(&real_time, &usr_time, &systm_time);
    brute_time = usr_time;
    FOR(j, inner_loops) {
      search_brute(&box);
    }
    ckit_get_times(&real_time, &usr_time, &systm_time);
    brute_time = usr_time - brute_time;
    cumulative_brute_time += brute_time;
    ckit_get_times(&real_time, &usr_time, &systm_time);
    lobs_time = usr_time;
    FOR(j, inner_loops) {
      search_lobs(&box);
    }
    ckit_get_times(&real_time, &usr_time, &systm_time);
    lobs_time = usr_time - lobs_time;
    cumulative_lobs_time += lobs_time;
  }

  fprintf (stderr, "Time per search:\n");
  fprintf (stderr, "brute force %g sec\n",
           cumulative_brute_time / (nsearches * inner_loops));
  fprintf (stderr, "LOBS %g sec (%g%%)\n",
           cumulative_lobs_time / (nsearches * inner_loops),
           100 * cumulative_lobs_time / cumulative_brute_time);
}




static void
wrong(int dim, int i, G3D_box_t *box)
{
  switch (dim) {
#if   TEST_DIM == 1
  case 1:
    fprintf(stderr, "box = {{%g}, {%g}}\n",
            box->lower.x, 
            box->upper.x);
    fprintf(stderr, "objbox = {{%g}, {%g}}\n",
            objs[i].box.lower.x, 
            objs[i].box.upper.x);
    fprintf(stderr, "1d status: brute = %d, lobs = %d\n",
            objs[i].found_by_brute_1d,
            objs[i].found_by_lobs_1d);
    lobs_1d_search (lobs1d,
                  box->lower.x,
                  box->upper.x);
    break;
#elif TEST_DIM == 2
  case 2:
    fprintf(stderr, "box = {{%g, %g}, {%g, %g}}\n",
            box->lower.x, box->lower.y, 
            box->upper.x, box->upper.y);
    fprintf(stderr, "objbox = {{%g, %g}, {%g, %g}}\n",
            objs[i].box.lower.x, objs[i].box.lower.y,
            objs[i].box.upper.x, objs[i].box.upper.y);
    fprintf(stderr, "2d status: brute = %d, lobs = %d\n",
            objs[i].found_by_brute_2d,
            objs[i].found_by_lobs_2d);
    lobs_2d_search (lobs2d,
                  box->lower.x,
                  box->lower.y,
                  box->upper.x,
                  box->upper.y);
    break;
#elif TEST_DIM == 3
  case 3:
    fprintf(stderr, "box = {{%g, %g, %g}, {%g, %g, %g}}\n",
            box->lower.x, box->lower.y, box->lower.z,
            box->upper.x, box->upper.y, box->upper.z);
    fprintf(stderr, "objbox = {{%g, %g, %g}, {%g, %g, %g}}\n",
            objs[i].box.lower.x, objs[i].box.lower.y, objs[i].box.lower.z,
            objs[i].box.upper.x, objs[i].box.upper.y, objs[i].box.upper.z);
    fprintf(stderr, "3d status: brute = %d, lobs = %d\n",
            objs[i].found_by_brute_3d,
            objs[i].found_by_lobs_3d);
    lobs_3d_search (lobs3d,
                  box->lower.x,
                  box->lower.y,
                  box->lower.z,
                  box->upper.x,
                  box->upper.y,
                  box->upper.z);
    break;
#endif
  default:
    errexit("unknown dimension");
    break;
  }
  errexit("disagreement");
}


static void
compare(G3D_box_t *box)
{
  int i;

  FOR(i, nobjs) {
#if   TEST_DIM == 1
    if (objs[i].found_by_lobs_1d != objs[i].found_by_brute_1d) {
      wrong(1, i, box);
    }
#elif TEST_DIM == 2
    if (objs[i].found_by_lobs_2d != objs[i].found_by_brute_2d) {
      wrong(2, i, box);
    }
#elif TEST_DIM == 3
    if (objs[i].found_by_lobs_3d != objs[i].found_by_brute_3d) {
      wrong(3, i, box);
    }
#endif
  }
}



#if TEST_BOX
static void
search_brute(G3D_box_t *box)
{
  int i;

  FOR(i, nobjs) {
#   if   TEST_DIM == 1
    objs[i].found_by_brute_1d = 0;
#   elif TEST_DIM == 2
    objs[i].found_by_brute_2d = 0;
#   elif TEST_DIM == 3
    objs[i].found_by_brute_3d = 0;
#   endif

#   if   TEST_DIM == 1
    if (G3D_RANGES_OVERLAP(*box, objs[i].box, x)) {
      objs[i].found_by_brute_1d = 1;
    }
#   elif TEST_DIM == 2
    if (G3D_RANGES_OVERLAP(*box, objs[i].box, x)) {
      if (G3D_RANGES_OVERLAP(*box, objs[i].box, y)) {
        objs[i].found_by_brute_2d = 1;
      }
    } 
#   elif TEST_DIM == 3
    if (G3D_RANGES_OVERLAP(*box, objs[i].box, x)) {
      if (G3D_RANGES_OVERLAP(*box, objs[i].box, y)) {
        if (G3D_RANGES_OVERLAP(*box, objs[i].box, z)) {
          objs[i].found_by_brute_3d = 1;
        }
      }
    } 
#   endif
  }
}
#else
static void
search_brute(G3D_box_t *box)
{
  int i;
  G3D_loc_t pt = box->lower;

  FOR(i, nobjs) {
#   if   TEST_DIM == 1
    objs[i].found_by_brute_1d = 0;
#   elif TEST_DIM == 2
    objs[i].found_by_brute_2d = 0;
#   elif TEST_DIM == 3
    objs[i].found_by_brute_3d = 0;
#   endif

#   if   TEST_DIM == 1
    if (G3D_VALUE_IN_RANGE(objs[i].box, x, pt.x)) {
      objs[i].found_by_brute_1d = 1;
    }
#   elif TEST_DIM == 2
    if (G3D_VALUE_IN_RANGE (objs[i].box, x, pt.x)) {
      if (G3D_VALUE_IN_RANGE (objs[i].box, y, pt.y)) {
        objs[i].found_by_brute_2d = 1;
      }
    } 
#   elif TEST_DIM == 3
    if (G3D_VALUE_IN_RANGE (objs[i].box, x, pt.x)) {
      if (G3D_VALUE_IN_RANGE (objs[i].box, y, pt.y)) {
        if (G3D_VALUE_IN_RANGE (objs[i].box, z, pt.z)) {
          objs[i].found_by_brute_3d = 1;
        }
      }
    } 
#   endif
  }
}
#endif





static void
search_lobs(G3D_box_t *box)
{
  obj *o;
  int i;

#if   TEST_DIM == 1
  lobs_1d_search (lobs1d,
                  box->lower.x,
                  box->upper.x);
#if !TIMING_MODE
  o = lobs_1d_first_overlapping_object (lobs1d);
  while (o) {
    o->found_by_lobs_1d++;
    o = lobs_1d_next_overlapping_object (lobs1d);
  }
#endif
  
#elif TEST_DIM == 2
  lobs_2d_search (lobs2d,
                  box->lower.x,
                  box->lower.y,
                  box->upper.x,
                  box->upper.y);
#if !TIMING_MODE
  o = lobs_2d_first_overlapping_object (lobs2d);
  while (o) {
    o->found_by_lobs_2d++;
    o = lobs_2d_next_overlapping_object (lobs2d);
  }
#endif
  
#elif TEST_DIM == 3
  lobs_3d_search (lobs3d,
                  box->lower.x,
                  box->lower.y,
                  box->lower.z,
                  box->upper.x,
                  box->upper.y,
                  box->upper.z);
#if !TIMING_MODE
  o = lobs_3d_first_overlapping_object (lobs3d);
  while (o) {
    o->found_by_lobs_3d++;
    o = lobs_3d_next_overlapping_object (lobs3d);
  }
#endif

#endif
}




static void 
process_options(int argc, char **argv)
{
  GET_OPT_ARG(argc, argv, "-nobjs", int, nobjs, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-nsearches", int, nsearches, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-per_leaf", int, per_leaf, {;},
              "number");
}


