#include <stdio.h>
#include <time.h>

#include "randP.h"
#include "timeP.h"
#include "getoptP.h"
#include "g3d.h"
#include "pirat.h"



#define NOBJS    1000
#define NSRCHS   5000

#include "lobs3d.h"


typedef struct obj {
  int        found;
  G3D_loc_t  loc;
  G3D_box_t  box;
}              obj;

static double xscalelo = -1;
static double xscalehi = 1;

static double yscalelo = 0.33333;
static double yscalehi = 0.9;

static double zscalelo = 0; 
static double zscalehi = 1; 


static int nobjs       = NOBJS;
static int nsearches   = NSRCHS;
static int pr_scale    = NSRCHS / 10;
static int per_leaf    = 10;
static int inner_loops = 10;
static double bsize    = 0.009;


static obj *objs;


static double cumulative_brute_time = 0;
static double cumulative_pirat_time = 0;
static double cumulative_lobs_time = 0;


static void 
generate(void);
static void
search_brute(G3D_box_t *box);
static void
search_lobs(G3D_box_t *box);
static void
search_pirat(G3D_box_t *box);
static void
time_it(void);
static void 
process_options(int argc, char **argv);


int  
main(int argc, char **argv)
{
  process_options (argc, argv);
  
  fprintf (stderr, "Timing LOBS vs. PiRaT: %d objects, %d searches\n",
           nobjs, nsearches * inner_loops);
  
  generate ();

  time_it ();

  fprintf (stderr, "Success\n");

  return 0;
}

#define LOBS_NEW_ARRAY(NELT, TYPE) \
                      ((TYPE *) calloc((NELT), (size_t) sizeof(TYPE)))


static int
errexit(char *s)
{
  fprintf(stderr, "Error: %s\n", s);
  exit(1);
  return 1;
}


static lobs_3d_front_end_t  *lobs3d = NULL;
static pirat_front_end_t *pirat = NULL;


/* Ordinary loop */
#define FOR(_Counter_, _Up_)                                      \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)

static void
next_obj(int i);



static void
generate(void)
{
  int i;

  objs = LOBS_NEW_ARRAY(nobjs, obj);
  if (objs == NULL) errexit("malloc failed");

  lobs3d = lobs_3d_new_instance ();
  lobs3d->max_num_of_objects_per_leaf = per_leaf;

  pirat = pirat_new_pirat (per_leaf, 0.0000001);
  
  FOR(i, nobjs) {
    next_obj(i);
    lobs_3d_add_object (lobs3d, &objs[i],
                        objs[i].box.lower.x,
                        objs[i].box.lower.y,
                        objs[i].box.lower.z,
                        objs[i].box.upper.x,
                        objs[i].box.upper.y,
                        objs[i].box.upper.z);
    if (!pirat_add_pt (pirat, &objs[i], &objs[i].loc)) {
      errexit ("failed pirat_add_pt");
    }
  }
}


static void
next_obj(int i)
{
  objs[i].loc.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
  objs[i].loc.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
  objs[i].loc.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
  G3D_INIT_BOX(objs[i].box, objs[i].loc);
  G3D_INFLATE_BOX(objs[i].box, bsize);
}



  


static void
time_it (void)
{
  int i, j;
  G3D_loc_t p;
  G3D_box_t box;
  double real_time, usr_time, systm_time;
  double brute_time, lobs_time, pirat_time;

  cumulative_pirat_time = 0;
  cumulative_lobs_time = 0;
  FOR(i, nsearches) {
    p.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
    p.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
    p.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
    G3D_INIT_BOX(box, p);

    ckit_get_times(&real_time, &usr_time, &systm_time);
    brute_time = usr_time;
    FOR(j, inner_loops) {
      search_brute (&box);
    }
    ckit_get_times(&real_time, &usr_time, &systm_time);
    brute_time = usr_time - brute_time;
    cumulative_brute_time += brute_time;

    ckit_get_times(&real_time, &usr_time, &systm_time);
    pirat_time = usr_time;
    FOR(j, inner_loops) {
      search_pirat (&box);
    }
    ckit_get_times(&real_time, &usr_time, &systm_time);
    pirat_time = usr_time - pirat_time;
    cumulative_pirat_time += pirat_time;

    ckit_get_times(&real_time, &usr_time, &systm_time);
    lobs_time = usr_time;
    FOR(j, inner_loops) {
      search_lobs(&box);
    }
    ckit_get_times(&real_time, &usr_time, &systm_time);
    lobs_time = usr_time - lobs_time;
    cumulative_lobs_time += lobs_time;
  }

  fprintf (stderr, "Time per search:\n");
  fprintf (stderr, "  brute force %g sec\n",
           cumulative_brute_time / (nsearches * inner_loops));
  fprintf (stderr, "  PiRaT %g sec (%g%%)\n",
           cumulative_pirat_time / (nsearches * inner_loops),
           100 * cumulative_pirat_time / cumulative_brute_time);
  fprintf (stderr, "  LOBS %g sec (%g%%)\n",
           cumulative_lobs_time / (nsearches * inner_loops),
           100 * cumulative_lobs_time / cumulative_brute_time);
}





static void
search_brute(G3D_box_t *box)
{
  int i;

  FOR(i, nobjs) {
    objs[i].found = 0;
    if (G3D_RANGES_OVERLAP(*box, objs[i].box, x)) {
      if (G3D_RANGES_OVERLAP(*box, objs[i].box, y)) {
        if (G3D_RANGES_OVERLAP(*box, objs[i].box, z)) {
          objs[i].found = 1;
        }
      }
    } 
  }
}





static void
search_lobs(G3D_box_t *box)
{
  obj *o;
  int i;

#if 0
  lobs_3d_search (lobs3d,
                  box->lower.x,
                  box->lower.y,
                  box->lower.z,
                  box->upper.x,
                  box->upper.y,
                  box->upper.z);
#else
  lobs_3d_search_for_boxes_overlapping_pt (lobs3d,
                                           box->lower.x,
                                           box->lower.y,
                                           box->lower.z);
#endif
}




static void
search_pirat (G3D_box_t *box)
{
  G3D_INFLATE_BOX (*box, bsize);
  
  pirat_search_range (pirat, box);
}


static void 
process_options(int argc, char **argv)
{
  GET_OPT_ARG(argc, argv, "-nobjs", int, nobjs, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-nsearches", int, nsearches, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-per_leaf", int, per_leaf, {;},
              "number");
}


